package de.meinefirma.meinprojekt.itt;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Der IttRunner ("Integration Test Tool Runner") fuehrt zu Input-CSV-Dateien die darin angegebenen Tests aus.
 * <pre>
 * Die Input-CSV-Dateien muessen Folgendes definieren:
 *    IttExecutionClass=...: Die das Interface IttExecutionInterf implementierende Java-Klasse,
 *                           welche die Parameter entgegennimmt, den Test ausfuehrt und die Ergebnisse returniert.
 *    Titelzeile mit den Semikolon-separierten Namen der Parameter.
 *    Beliebig viele Zeilen mit den Semikolon-separierten Parametern.
 * Beispiel:
 *    // Beliebiger Kommentar, der nicht in HTML-Output angezeigt wird
 *    ## Beliebiger Kommentar, der nicht in HTML-Output angezeigt wird
 *    Kommentar=Dieser Kommentar wird in HTML-Output angezeigt
 *    IttExecutionClass=de.meinefirma.meinprojekt.itt.MeineIttExecutionClass
 *    Key1; Key2; Key3
 *       1;    2;    3
 *       4;    5;    9
 * </pre>
 */
public class IttRunner
{
   private static final String INPUT_CSV_DATEI_POSTFIX = "ITT.csv";

   /**
    * Fuehre zu allen gefundenen Input-CSV-Dateien die darin angegebenen Tests aus.
    * @param args Als Parameter kann ein Datei-Postfix angegeben werden.
    *             Es werden Input-CSV-Dateien verwendet, deren Name mit diesem Postfix enden.
    *             Falls null oder ein Leerstring uebergeben wird, wird der Default "ITT.csv" verwendet.
    *             Die Input-CSV-Dateien werden im aktuellen Verzeichnispfad und in den Classpath-Pfaden gesucht.
    * @exit() mit 0 falls erfolgreich und keine Fehler, sonst mit 1.
    */
   public static void main( String[] args ) throws IOException
   {
      System.exit( ( runAllTests( ( args.length > 0 ) ? args[0] : null, true, true ) ) ? 0 : 1 );
   }

   /**
    * Fuehre zu allen gefundenen Input-CSV-Dateien die darin angegebenen Tests aus.
    * @param inputCsvDateiPostfix Es werden Input-CSV-Dateien verwendet, deren Name mit diesem Postfix enden.
    *                             Falls null oder ein Leerstring uebergeben wird, wird der Default "ITT.csv" verwendet.
    *                             Die Input-CSV-Dateien werden im aktuellen Verzeichnispfad und in den Classpath-Pfaden gesucht.
    * @param showErrMsg           true falls Fehlermeldungen in der Konsole gewuenscht sind.
    * @param showHtmlResult       true falls Webbrowser mit Ergebnis-HTML-Seite aufgerufen werden soll.
    * @return true falls erfolgreich und keine Fehler.
    */
   public static boolean runAllTests( String inputCsvDateiPostfix, boolean showErrMsg, boolean showHtmlResult ) throws IOException
   {
      String csvPostfix = ( inputCsvDateiPostfix != null && inputCsvDateiPostfix.trim().length() > 0 )
                          ? inputCsvDateiPostfix.trim() : INPUT_CSV_DATEI_POSTFIX;
      List<File> inputCsvFiles = findFiles( csvPostfix );
      if( inputCsvFiles == null || inputCsvFiles.size() == 0 ) {
         System.out.println( "Fehler: Keine Input-CSV-Datei mit dem Postfix '" + csvPostfix + "' gefunden." );
         return false;
      }
      boolean resultOK = true;
      for( File inputCsvFile : inputCsvFiles ) {
         boolean b = runTest( inputCsvFile, null );
         resultOK &= b;
         if( !b && showErrMsg ) { System.out.println( "Es gab Fehler mit '" + inputCsvFile.getName() + "'." ); }
         if( showHtmlResult ) {
            Runtime.getRuntime().exec( "cmd /c start target/" + inputCsvFile.getName() + ".html" );
            try { Thread.sleep( 1500 ); } catch( InterruptedException e ) {/*ok*/}
         }
      }
      return resultOK;
   }

   /**
    * Fuehre zu einzelner Input-CSV-Datei die darin angegebenen Tests aus.
    * @param inputCsvFile   Input-CSV-Datei mit Tests.
    * @param outputHtmlFile Output-HTML-Datei fuer das Ergebnis.
    * @return true falls erfolgreich und keine Fehler.
    */
   public static boolean runTest( File inputCsvFile, String outputHtmlFile )
   {
      String outFile = ( outputHtmlFile != null ) ? outputHtmlFile : "target/" + inputCsvFile.getName() + ".html";
      boolean resultOK = true;
      BufferedReader in  = null;
      BufferedWriter out = null;
      try {
         in  = new BufferedReader( new InputStreamReader( new FileInputStream( inputCsvFile ) ) );
         out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( outFile ) ) );
         Class<?> ittExecutionClass = null;
         String   kommentar = "";
         List<String> keys = null;
         String line;
         while( (line = in.readLine()) != null ) {
            if( line.startsWith( "Kommentar=" ) ) {
               kommentar += line.substring( line.indexOf( '=' ) + 1 ).trim();
            }
            if( line.startsWith( "IttExecutionClass=" ) ) {
               String iec = line.substring( line.indexOf( '=' ) + 1 ).trim();
               if( iec.length() > 0 ) {
                  ittExecutionClass = Class.forName( iec );
                  if( ittExecutionClass.newInstance() instanceof IttExecutionInterf ) {
                     break;
                  }
                  System.out.println( "Warnung: Die Klasse '" + ittExecutionClass + "' zum Eintrag 'IttExecutionClass=...' in der Input-CSV-Datei '" + inputCsvFile.getName() + "' implementiert nicht das Interface 'IttExecutionInterf'." );
                  ittExecutionClass = null;
               }
            }
         }
         if( ittExecutionClass == null ) {
            System.out.println( "Fehler: Keinen gueltigen Eintrag 'IttExecutionClass=...' in der Input-CSV-Datei '" + inputCsvFile.getName() + "' gefunden." );
            resultOK = false;
            return false;
         }
         out.write( "<html>\n" +
         		"<head>\n" +
         		"  <title>Integration Test Tool (\"ITT\")</title>\n" +
         		"  <meta http-equiv='content-type' content='text/html;charset=iso-8859-1'>\n" +
         		"  <style type='text/css'>\n" +
         		"  <!--\n" +
         		"    body, th, td { font-family   : Calibri,Arial,Helvetica,sans-serif }\n" +
         		"    th           { font-size     : 0.9em; }\n" +
         		"    td           { font-size     : 0.8em; }\n" +
         		"    h1, h2       { margin-top    : 10.0pt;\n" +
         		"                   margin-bottom : 10.0pt; }\n" +
         		"    a:hover      { color         : green }\n" +
         		"  -->\n" +
         		"  </style>\n" +
         		"</head>\n" );
         out.write( "<h3>Integration Test Tool (\"ITT\")</h3>\n" );
         out.write( "<table>\n" );
         out.write( "<tr><td>Input-CSV-Datei:  </td><td>&nbsp;</td><td>" + inputCsvFile + "</td></tr>\n" );
         out.write( "<tr><td>Output-HTML-Datei:</td><td>&nbsp;</td><td>" + outFile + "</td></tr>\n" );
         out.write( "<tr><td>IttExecutionClass:</td><td>&nbsp;</td><td>" + ittExecutionClass.getName() + "</td></tr>\n" );
         if( kommentar.trim().length() > 0 ) {
            out.write( "<tr><td>Kommentar:        </td><td>&nbsp;</td><td>" + kommentar + "</td></tr>\n" );
         }
         out.write( "</table><br>\n" );
         out.write( "<table border='1' cellspacing='0' cellpadding='2'>\n" );
         while( (line = in.readLine()) != null ) {
            if( line.startsWith( "//" ) || line.startsWith( "##" ) || !line.contains( ";" ) ) { continue; }
            List<String> values = Arrays.asList( line.split( ";" ) );
            if( values == null || values.size() == 0 ) { continue; }
            for( int i = 0; i < values.size(); i++ ) {
               String s = values.get( i );
               if( s != null ) { values.set( i, s.trim() ); }
            }
            if( keys == null ) {
               keys = values;
               StringBuilder sb = new StringBuilder();
               sb.append( "<tr bgcolor='#EBEEEE'>" );
               for( String k : keys ) {
                  sb.append( "<th>" ).append( k ).append( "</th>" );
               }
               sb.append( "</tr>\n" );
               out.write( sb.toString() );
            } else {
               List<IttTestParm> testParmList = new ArrayList<IttTestParm>();
               for( int i = 0; i < keys.size(); i++ ) {
                  testParmList.add( new IttTestParm( keys.get( i ), ( i < values.size() ) ? values.get( i ) : null ) );
               }
               try {
                  resultOK &= runTestLine( ittExecutionClass, testParmList, out );
               } catch( Exception ex ) {
                  System.out.println( "Fehler mit Input-CSV-Datei '" + inputCsvFile + "' und der Zeile '" + line + "':\n" + ex );
                  resultOK = false;
               }
            }
         }
         out.write( "</table>\n" );
         out.write( "<p>" + "Ergebnis: " + (( resultOK ) ? "OK" : "Fehler") + "</p>\n" );
         out.write( "</html>\n" );
      } catch( Exception ex ) {
         System.out.println( "Fehler mit Input-CSV-Datei '" + inputCsvFile + "' oder Output-HTML-Datei '" + outFile + "':\n" + ex );
         resultOK = false;
      } finally {
         if( in  != null ) { try { in.close(); } catch( IOException ex ) {/*ok*/} }
         if( out != null ) { try { out.close(); } catch( IOException ex ) {/*ok*/} }
      }
      return resultOK;
   }

   private static boolean runTestLine( Class<?> ittExecutionClass, List<IttTestParm> testParmList, BufferedWriter out ) throws IOException
   {
      boolean resultOK = true;
      String exeptionMsg = null;
      if( ittExecutionClass == null ) {
         exeptionMsg = "Fehler: Keinen Eintrag 'IttExecutionClass=...' in der Input-CSV-Datei gefunden.";
         System.out.println( exeptionMsg );
         return false;
      }
      try {
         IttExecutionInterf ittExecution = (IttExecutionInterf) ittExecutionClass.newInstance();
         ittExecution.execute( transformListToMap( testParmList ) );
      } catch( Exception ex ) {
         exeptionMsg = ex.toString();
         resultOK = false;
      }
      StringBuilder sb = new StringBuilder();
      sb.append( "<tr>" );
      for( IttTestParm tp : testParmList ) {
         boolean resultStateOK  = (tp.resultState != null &&  tp.resultState.booleanValue())
                                  || (tp.resultValue != null && isEqual( tp.value, tp.resultValue ));
         boolean resultStateErr = (tp.resultState != null && !tp.resultState.booleanValue())
                                  || (tp.resultValue != null && !isEqual( tp.value, tp.resultValue ))
                                  || tp.errMsg != null || exeptionMsg != null;
         resultOK &= !resultStateErr;
         String s = ( tp.value == null && !"Bemerkung".equals( tp.key ) ) ? "{null}" : tp.value;
         if( tp.resultValue != null && !isEqual( tp.value, tp.resultValue ) ) {
            s += " {erwartet};<br>" + tp.resultValue + " {gefunden}";
         }
         if( tp.errMsg != null ) { s += "<br>" + tp.errMsg; }
         if( exeptionMsg != null) { s = "Exception: " + exeptionMsg; }
         String bgcolor = ( resultStateErr ) ? " bgcolor='red'" : ( resultStateOK ) ? " bgcolor='lime'" : "";
         if( s == null || s.trim().length() == 0 ) { s = "&nbsp;"; }
         sb.append( "<td" ).append( bgcolor ).append( ">" ).append( s );
         sb.append( "</td>" );
      }
      sb.append( "</tr>\n" );
      out.write( sb.toString() );
      return resultOK;
   }

   private static Map<String,IttTestParm> transformListToMap( List<IttTestParm> testParmList ) {
      Map<String,IttTestParm> map = new HashMap<String,IttTestParm>();
      for( IttTestParm tp : testParmList ) {
         map.put( tp.key, tp );
      }
      return map;
   }

   private static List<File> findFilesInDir( final String dir, final String nameEndsWith )
   {
      List<File> result = new ArrayList<File>();
      File[] ff = new File( dir ).listFiles( new FilenameFilter() {
         @Override public boolean accept( File verz, String name ) {
            return name.endsWith( nameEndsWith );
         }
      });
      if( ff != null ) { result.addAll( Arrays.asList( ff ) ); }
      Collections.sort( result );
      return result;
   }

   private static List<File> findFiles( final String nameEndsWith )
   {
      List<File>   result = new ArrayList<File>();
      List<String> pp = getPathesFromClasspath();
      pp.add( 0, "." );
      for( String p : pp ) { result.addAll( findFilesInDir( p, nameEndsWith ) ); }
      Collections.sort( result );
      return result;
   }

   private static List<String> getPathesFromClasspath()
   {
      String          classpath     = System.getProperty( "java.class.path" );
      String          pathseparator = System.getProperty( "path.separator" );
      StringTokenizer tokenizer     = new StringTokenizer( classpath, pathseparator );
      List<String>    pathes        = new ArrayList<String>();
      while( tokenizer.hasMoreElements() ) { pathes.add( tokenizer.nextToken() ); }
      return pathes;
   }

   private static boolean isEqual( String s1, String s2 )
   {
      return (s1 == null && s2 == null) || (s1 != null && s1.equals( s2 ));
   }
}
